@@ -23,8 +23,8 @@ from ray.rllib.models.torch.torch_action_dist import (
 from ray.rllib.utils.framework import try_import_torch
 from ray.rllib.utils.spaces.simplex import Simplex
 from ray.rllib.utils.torch_ops import apply_grad_clipping, huber_loss
-from ray.rllib.utils.typing import LocalOptimizer, TensorType, \
-    TrainerConfigDict
+from ray.rllib.utils.typing import LocalOptimizer, ModelInputDict, \
+    TensorType, TrainerConfigDict
 
 torch, nn = try_import_torch()
 F = nn.functional
@@ -32,6 +32,29 @@ F = nn.functional
 logger = logging.getLogger(__name__)
 
 
+def _get_dist_class(config: TrainerConfigDict, action_space: gym.spaces.Space
+                    ) -> Type[TorchDistributionWrapper]:
+    """Helper function to return a dist class based on config and action space.
+
+    Args:
+        config (TrainerConfigDict): The Trainer's config dict.
+        action_space (gym.spaces.Space): The action space used.
+
+    Returns:
+        Type[TFActionDistribution]: A TF distribution class.
+    """
+    if isinstance(action_space, Discrete):
+        return TorchCategorical
+    elif isinstance(action_space, Simplex):
+        return TorchDirichlet
+    else:
+        if config["normalize_actions"]:
+            return TorchSquashedGaussian if \
+                not config["_use_beta_distribution"] else TorchBeta
+        else:
+            return TorchDiagGaussian
+
+
 def build_sac_model_and_action_dist(
         policy: Policy,
         obs_space: gym.spaces.Space,
@@ -56,33 +79,10 @@ def build_sac_model_and_action_dist(
     return model, action_dist_class
 
 
-def _get_dist_class(config: TrainerConfigDict, action_space: gym.spaces.Space
-                    ) -> Type[TorchDistributionWrapper]:
-    """Helper function to return a dist class based on config and action space.
-
-    Args:
-        config (TrainerConfigDict): The Trainer's config dict.
-        action_space (gym.spaces.Space): The action space used.
-
-    Returns:
-        Type[TFActionDistribution]: A TF distribution class.
-    """
-    if isinstance(action_space, Discrete):
-        return TorchCategorical
-    elif isinstance(action_space, Simplex):
-        return TorchDirichlet
-    else:
-        if config["normalize_actions"]:
-            return TorchSquashedGaussian if \
-                not config["_use_beta_distribution"] else TorchBeta
-        else:
-            return TorchDiagGaussian
-
-
 def action_distribution_fn(
         policy: Policy,
         model: ModelV2,
-        obs_batch: TensorType,
+        input_dict: ModelInputDict,
         *,
         state_batches: Optional[List[TensorType]] = None,
         seq_lens: Optional[TensorType] = None,
@@ -106,8 +106,8 @@ def action_distribution_fn(
         model (TorchModelV2): The SAC specific Model to use to generate the
             distribution inputs (see sac_tf|torch_model.py). Must support the
             `get_policy_output` method.
-        obs_batch (TensorType): The observations to be used as inputs to the
-            model.
+        input_dict (ModelInputDict): The input-dict to be used for the model
+            call.
         state_batches (Optional[List[TensorType]]): The list of internal state
             tensor batches.
         seq_lens (Optional[TensorType]): The tensor of sequence lengths used
@@ -127,10 +127,7 @@ def action_distribution_fn(
             (in the RNN case).
     """
     # Get base-model output (w/o the SAC specific parts of the network).
-    model_out, _ = model({
-        "obs": obs_batch,
-        "is_training": is_training,
-    }, [], None)
+    model_out, _ = model(input_dict, [], None)
     # Use the base output to get the policy outputs from the SAC model's
     # policy components.
     distribution_inputs = model.get_policy_output(model_out)
@@ -304,7 +301,12 @@ def actor_critic_loss(
     policy.q_t = q_t
     policy.policy_t = policy_t
     policy.log_pis_t = log_pis_t
-    policy.td_error = td_error
+
+    # Store td-error in model, such that for multi-GPU, we do not override
+    # them during the parallel loss phase. TD-error tensor in final stats
+    # can then be concatenated and retrieved for each individual batch item.
+    model.td_error = td_error
+
     policy.actor_loss = actor_loss
     policy.critic_loss = critic_loss
     policy.alpha_loss = alpha_loss
@@ -327,9 +329,15 @@ def stats(policy: Policy, train_batch: SampleBatch) -> Dict[str, TensorType]:
     Returns:
         Dict[str, TensorType]: The stats dict.
     """
+    td_error = torch.cat(
+        [
+            getattr(t, "td_error", torch.tensor([0.0]))
+            for t in policy.model_gpu_towers
+        ],
+        dim=0)
     return {
-        "td_error": policy.td_error,
-        "mean_td_error": torch.mean(policy.td_error),
+        "td_error": td_error,
+        "mean_td_error": torch.mean(td_error),
         "actor_loss": torch.mean(policy.actor_loss),
         "critic_loss": torch.mean(torch.stack(policy.critic_loss)),
         "alpha_loss": torch.mean(policy.alpha_loss),
