@@ -12,24 +12,27 @@ import shelve
 
 import ray
 import ray.cloudpickle as cloudpickle
+from ray.rllib.agents.registry import get_trainer_class
 from ray.rllib.env import MultiAgentEnv
 from ray.rllib.env.base_env import _DUMMY_AGENT_ID
 from ray.rllib.env.env_context import EnvContext
 from ray.rllib.evaluation.worker_set import WorkerSet
 from ray.rllib.policy.sample_batch import DEFAULT_POLICY_ID
-from ray.rllib.utils.deprecation import deprecation_warning
 from ray.rllib.utils.spaces.space_utils import flatten_to_single_ndarray
 from ray.tune.utils import merge_dicts
 from ray.tune.registry import get_trainable_cls, _global_registry, ENV_CREATOR
 
 EXAMPLE_USAGE = """
-Example Usage via RLlib CLI:
+Example usage via RLlib CLI:
     rllib rollout /tmp/ray/checkpoint_dir/checkpoint-0 --run DQN
     --env CartPole-v0 --steps 1000000 --out rollouts.pkl
 
-Example Usage via executable:
+Example usage via executable:
     ./rollout.py /tmp/ray/checkpoint_dir/checkpoint-0 --run DQN
     --env CartPole-v0 --steps 1000000 --out rollouts.pkl
+
+Example usage w/o checkpoint (for testing purposes):
+    ./rollout.py --run PPO --env CartPole-v0 --episodes 500
 """
 
 # Note: if you use any custom models or envs, register them here first, e.g.:
@@ -42,6 +45,94 @@ Example Usage via executable:
 # register_env("pa_cartpole", lambda _: ParametricActionsCartPole(10))
 
 
+def create_parser(parser_creator=None):
+    parser_creator = parser_creator or argparse.ArgumentParser
+    parser = parser_creator(
+        formatter_class=argparse.RawDescriptionHelpFormatter,
+        description="Roll out a reinforcement learning agent "
+        "given a checkpoint.",
+        epilog=EXAMPLE_USAGE)
+
+    parser.add_argument(
+        "checkpoint",
+        type=str,
+        nargs="?",
+        help="(Optional) checkpoint from which to roll out. "
+        "If none given, will use an initial (untrained) Trainer.")
+
+    required_named = parser.add_argument_group("required named arguments")
+    required_named.add_argument(
+        "--run",
+        type=str,
+        required=True,
+        help="The algorithm or model to train. This may refer to the name "
+        "of a built-on algorithm (e.g. RLLib's `DQN` or `PPO`), or a "
+        "user-defined trainable function or class registered in the "
+        "tune registry.")
+    required_named.add_argument(
+        "--env",
+        type=str,
+        help="The environment specifier to use. This could be an openAI gym "
+        "specifier (e.g. `CartPole-v0`) or a full class-path (e.g. "
+        "`ray.rllib.examples.env.simple_corridor.SimpleCorridor`).")
+    parser.add_argument(
+        "--local-mode",
+        action="store_true",
+        help="Run ray in local mode for easier debugging.")
+    parser.add_argument(
+        "--no-render",
+        default=False,
+        action="store_const",
+        const=True,
+        help="Suppress rendering of the environment.")
+    parser.add_argument(
+        "--video-dir",
+        type=str,
+        default=None,
+        help="Specifies the directory into which videos of all episode "
+        "rollouts will be stored.")
+    parser.add_argument(
+        "--steps",
+        default=10000,
+        help="Number of timesteps to roll out. Rollout will also stop if "
+        "`--episodes` limit is reached first. A value of 0 means no "
+        "limitation on the number of timesteps run.")
+    parser.add_argument(
+        "--episodes",
+        default=0,
+        help="Number of complete episodes to roll out. Rollout will also stop "
+        "if `--steps` (timesteps) limit is reached first. A value of 0 means "
+        "no limitation on the number of episodes run.")
+    parser.add_argument("--out", default=None, help="Output filename.")
+    parser.add_argument(
+        "--config",
+        default="{}",
+        type=json.loads,
+        help="Algorithm-specific configuration (e.g. env, hyperparams). "
+        "Gets merged with loaded configuration from checkpoint file and "
+        "`evaluation_config` settings therein.")
+    parser.add_argument(
+        "--save-info",
+        default=False,
+        action="store_true",
+        help="Save the info field generated by the step() method, "
+        "as well as the action, observations, rewards and done fields.")
+    parser.add_argument(
+        "--use-shelve",
+        default=False,
+        action="store_true",
+        help="Save rollouts into a python shelf file (will save each episode "
+        "as it is generated). An output filename must be set using --out.")
+    parser.add_argument(
+        "--track-progress",
+        default=False,
+        action="store_true",
+        help="Write progress to a temporary file (updated "
+        "after each episode). An output filename must be set using --out; "
+        "the progress file will live in the same folder.")
+    return parser
+
+
 class RolloutSaver:
     """Utility class for storing rollouts.
 
@@ -165,108 +256,31 @@ class RolloutSaver:
         self._total_steps += 1
 
 
-def create_parser(parser_creator=None):
-    parser_creator = parser_creator or argparse.ArgumentParser
-    parser = parser_creator(
-        formatter_class=argparse.RawDescriptionHelpFormatter,
-        description="Roll out a reinforcement learning agent "
-        "given a checkpoint.",
-        epilog=EXAMPLE_USAGE)
-
-    parser.add_argument(
-        "checkpoint", type=str, help="Checkpoint from which to roll out.")
-    required_named = parser.add_argument_group("required named arguments")
-    required_named.add_argument(
-        "--run",
-        type=str,
-        required=True,
-        help="The algorithm or model to train. This may refer to the name "
-        "of a built-on algorithm (e.g. RLLib's DQN or PPO), or a "
-        "user-defined trainable function or class registered in the "
-        "tune registry.")
-    required_named.add_argument(
-        "--env", type=str, help="The gym environment to use.")
-    parser.add_argument(
-        "--no-render",
-        default=False,
-        action="store_const",
-        const=True,
-        help="Suppress rendering of the environment.")
-    parser.add_argument(
-        "--monitor",
-        default=False,
-        action="store_true",
-        help="Wrap environment in gym Monitor to record video. NOTE: This "
-        "option is deprecated: Use `--video-dir [some dir]` instead.")
-    parser.add_argument(
-        "--video-dir",
-        type=str,
-        default=None,
-        help="Specifies the directory into which videos of all episode "
-        "rollouts will be stored.")
-    parser.add_argument(
-        "--steps",
-        default=10000,
-        help="Number of timesteps to roll out (overwritten by --episodes).")
-    parser.add_argument(
-        "--episodes",
-        default=0,
-        help="Number of complete episodes to roll out (overrides --steps).")
-    parser.add_argument("--out", default=None, help="Output filename.")
-    parser.add_argument(
-        "--config",
-        default="{}",
-        type=json.loads,
-        help="Algorithm-specific configuration (e.g. env, hyperparams). "
-        "Gets merged with loaded configuration from checkpoint file and "
-        "`evaluation_config` settings therein.")
-    parser.add_argument(
-        "--save-info",
-        default=False,
-        action="store_true",
-        help="Save the info field generated by the step() method, "
-        "as well as the action, observations, rewards and done fields.")
-    parser.add_argument(
-        "--use-shelve",
-        default=False,
-        action="store_true",
-        help="Save rollouts into a python shelf file (will save each episode "
-        "as it is generated). An output filename must be set using --out.")
-    parser.add_argument(
-        "--track-progress",
-        default=False,
-        action="store_true",
-        help="Write progress to a temporary file (updated "
-        "after each episode). An output filename must be set using --out; "
-        "the progress file will live in the same folder.")
-    return parser
-
-
 def run(args, parser):
     # Load configuration from checkpoint file.
-    config_dir = os.path.dirname(args.checkpoint)
-    config_path = os.path.join(config_dir, "params.pkl")
-    # Try parent directory.
-    if not os.path.exists(config_path):
-        config_path = os.path.join(config_dir, "../params.pkl")
-
-    # If no pkl file found, require command line `--config`.
-    if not os.path.exists(config_path):
-        if not args.config:
-            raise ValueError(
-                "Could not find params.pkl in either the checkpoint dir or "
-                "its parent directory AND no config given on command line!")
-        else:
-            config = args.config
+    config_path = ""
+    if args.checkpoint:
+        config_dir = os.path.dirname(args.checkpoint)
+        config_path = os.path.join(config_dir, "params.pkl")
+        # Try parent directory.
+        if not os.path.exists(config_path):
+            config_path = os.path.join(config_dir, "../params.pkl")
 
     # Load the config from pickled.
-    else:
+    if os.path.exists(config_path):
         with open(config_path, "rb") as f:
             config = cloudpickle.load(f)
+    # If no pkl file found, require command line `--config`.
+    else:
+        # If no config in given checkpoint -> Error.
+        if args.checkpoint:
+            raise ValueError(
+                "Could not find params.pkl in either the checkpoint dir or "
+                "its parent directory AND no `--config` given on command "
+                "line!")
 
-    # Set num_workers to be at least 2.
-    if "num_workers" in config:
-        config["num_workers"] = min(2, config["num_workers"])
+        # Use default config for given agent.
+        _, config = get_trainer_class(args.run, return_config=True)
 
     # Make sure worker 0 has an Env.
     config["create_env_on_driver"] = True
@@ -285,25 +299,31 @@ def run(args, parser):
             parser.error("the following arguments are required: --env")
         args.env = config.get("env")
 
-    ray.init()
+    # Make sure we have evaluation workers.
+    if not config.get("evaluation_num_workers"):
+        config["evaluation_num_workers"] = config.get("num_workers", 0)
+    if not config.get("evaluation_num_episodes"):
+        config["evaluation_num_episodes"] = 1
+    config["render_env"] = not args.no_render
+    config["record_env"] = args.video_dir
+
+    ray.init(local_mode=args.local_mode)
 
     # Create the Trainer from config.
     cls = get_trainable_cls(args.run)
     agent = cls(env=args.env, config=config)
-    # Load state from checkpoint.
-    agent.restore(args.checkpoint)
+
+    # Load state from checkpoint, if provided.
+    if args.checkpoint:
+        agent.restore(args.checkpoint)
+
     num_steps = int(args.steps)
     num_episodes = int(args.episodes)
 
     # Determine the video output directory.
-    # Deprecated way: Use (--out|~/ray_results) + "/monitor" as dir.
     video_dir = None
-    if args.monitor:
-        video_dir = os.path.join(
-            os.path.dirname(args.out or "")
-            or os.path.expanduser("~/ray_results/"), "monitor")
-    # New way: Allow user to specify a video output path.
-    elif args.video_dir:
+    # Allow user to specify a video output path.
+    if args.video_dir:
         video_dir = os.path.expanduser(args.video_dir)
 
     # Do the actual rollout.
@@ -333,13 +353,13 @@ def default_policy_agent_mapping(unused_agent_id):
 
 def keep_going(steps, num_steps, episodes, num_episodes):
     """Determine whether we've collected enough data"""
-    # if num_episodes is set, this overrides num_steps
-    if num_episodes:
-        return episodes < num_episodes
-    # if num_steps is set, continue until we reach the limit
-    if num_steps:
-        return steps < num_steps
-    # otherwise keep going forever
+    # If num_episodes is set, stop if limit reached.
+    if num_episodes and episodes >= num_episodes:
+        return False
+    # If num_steps is set, stop if limit reached.
+    elif num_steps and steps >= num_steps:
+        return False
+    # Otherwise, keep going.
     return True
 
 
@@ -355,16 +375,36 @@ def rollout(agent,
     if saver is None:
         saver = RolloutSaver()
 
-    if hasattr(agent, "workers") and isinstance(agent.workers, WorkerSet):
+    # Normal case: Agent was setup correctly with an evaluation WorkerSet,
+    # which we will now use to rollout.
+    if hasattr(agent, "evaluation_workers") and isinstance(
+            agent.evaluation_workers, WorkerSet):
+        steps = 0
+        episodes = 0
+        while keep_going(steps, num_steps, episodes, num_episodes):
+            saver.begin_rollout()
+            eval_result = agent.evaluate()["evaluation"]
+            # Increase timestep and episode counters.
+            eps = agent.config["evaluation_num_episodes"]
+            episodes += eps
+            steps += eps * eval_result["episode_len_mean"]
+            # Print out results and continue.
+            print("Episode #{}: reward: {}".format(
+                episodes, eval_result["episode_reward_mean"]))
+            saver.end_rollout()
+        return
+
+    # Agent has no evaluation workers, but RolloutWorkers.
+    elif hasattr(agent, "workers") and isinstance(agent.workers, WorkerSet):
         env = agent.workers.local_worker().env
         multiagent = isinstance(env, MultiAgentEnv)
         if agent.workers.local_worker().multiagent:
             policy_agent_mapping = agent.config["multiagent"][
                 "policy_mapping_fn"]
-
         policy_map = agent.workers.local_worker().policy_map
         state_init = {p: m.get_initial_state() for p, m in policy_map.items()}
         use_lstm = {p: len(s) > 0 for p, s in state_init.items()}
+    # Agent has neither evaluation- nor rollout workers.
     else:
         from gym import envs
         if envs.registry.env_specs.get(agent.config["env"]):
@@ -397,7 +437,7 @@ def rollout(agent,
         env = gym_wrappers.Monitor(
             env=env,
             directory=video_dir,
-            video_callable=lambda x: True,
+            video_callable=lambda _: True,
             force=True)
 
     steps = 0
@@ -470,15 +510,6 @@ if __name__ == "__main__":
     parser = create_parser()
     args = parser.parse_args()
 
-    # Old option: monitor, use video-dir instead.
-    if args.monitor:
-        deprecation_warning("--monitor", "--video-dir=[some dir]")
-    # User tries to record videos, but no-render is set: Error.
-    if (args.monitor or args.video_dir) and args.no_render:
-        raise ValueError(
-            "You have --no-render set, but are trying to record rollout videos"
-            " (via options --video-dir/--monitor)! "
-            "Either unset --no-render or do not use --video-dir/--monitor.")
     # --use_shelve w/o --out option.
     if args.use_shelve and not args.out:
         raise ValueError(
